/**************************************************************************
 *
 * FILE  encoding.c
 * Copyright (c) 2015, 2017 Daniel Kahlin <daniel@kahlin.net>
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   optimize encoding and generate
 *
 ******/
#include <limits.h>
#include <stdio.h>
#include <stdint.h>

#include "bitfunc.h"
#include "encoding.h"
#include "global.h"
#include "match.h"
#include "message.h"
#include "utils.h"

#define DEBUG_COMPLEXITY 0

#define HAVE_SINGLE_BYTE 1
#define HAVE_THREE_BYTE 1

#if HAVE_SINGLE_BYTE
#define MIN_MATCH 1
#else
#define MIN_MATCH 2
#endif



PrimaryPath *create_primarypath(int n, uint8_t *buf)
{
    PrimaryPath *pp;

    pp = safe_malloc(sizeof(PrimaryPath), "primarypath");
    pp->path = safe_malloc(n * sizeof(Match), "matches");
    pp->n = n;
    pp->buf = buf;

    return pp;
}

void destroy_primarypath(PrimaryPath *pp)
{
    if (pp) {
	free(pp->path);
	free(pp);
    }
}


/**************************************************************************
 *
 * SECTION  fast cost functions
 *
 ******/
uint32_t litcost[0x10000];
uint32_t lencost[0x10000];
#if HAVE_SINGLE_BYTE
uint32_t offscost1[0x10000];
#endif
uint32_t offscost2[0x10000];
#if HAVE_THREE_BYTE
uint32_t offscost3[0x10000];
#endif
uint32_t offscost[0x10000];

static void prepare_fast(CostFuncSet *cfs, EncodingSet *es)
{
    int i;

    /* prepare tables for fast cost calculations */
    for (i = 0; i < 0x10000; i++) {
	litcost[i] = cfs->cost_lit(es, i);
	lencost[i] = cfs->cost_mlen(es, i);
#if HAVE_SINGLE_BYTE
	offscost1[i] = cfs->cost_moffs(es, i, 1);
#endif
	offscost2[i] = cfs->cost_moffs(es, i, 2);
#if HAVE_THREE_BYTE
	offscost3[i] = cfs->cost_moffs(es, i, 3);
	offscost[i] = cfs->cost_moffs(es, i, 4);
#else
	offscost[i] = cfs->cost_moffs(es, i, 3);
#endif
    }
}

static inline int fast_lit(int l)
{
    return litcost[l];
}

static inline int fast_mlen(int l)
{
    return lencost[l];
}

static inline int fast_moffs(int of, int l)
{
    switch (l) {
#if HAVE_SINGLE_BYTE
    case 1:
	return offscost1[of];
#endif
    case 2:
	return offscost2[of];
#if HAVE_THREE_BYTE
    case 3:
	return offscost3[of];
#endif
    default:
	return offscost[of];
    }
}


/**************************************************************************
 *
 * NAME  find_cheapest_path
 *
 * DESCRIPTION
 *   Consider encoding cost to find the cheapest path through all matches
 *
 ******/
PrimaryPath *find_cheapest_path(MatchTree *mt, CostFuncSet *cfs, EncodingSet *es)
{
    int i, j;
    int cur;
    int len = mt->len;
    int *dist;
    int *prev;
    Match *path;

    /* construct lookup tables */
    prepare_fast(cfs, es);

    /* create path tables */
    dist = safe_malloc(sizeof(int) * (len + 1), "dist table");
    prev = safe_malloc(sizeof(int) * (len + 1), "prev table");
    path = safe_malloc(sizeof(Match) * (len + 1), "match table");

    /* initialize costs */
    for (i = 0; i < len+1; i++) {
	dist[i] = INT_MAX;
	prev[i] = -1;
	//path[i] = 0;
    }
    dist[0] = 0;

#if DEBUG_COMPLEXITY
    int n_match, n_match_taken, n_rle, n_rle_taken, n_lt, n_lt_taken;
    n_match = 0;
    n_match_taken = 0;
    n_rle = 0;
    n_rle_taken = 0;
    n_lt = 0;
    n_lt_taken = 0;
#endif

    /*
     * calculate costs
     */
    cur = 0;
    while (cur < len) {
	int w, v;
	Match *m = mt->match[cur];

	while (m && !is_end(m)) {

	    if (is_match(m) && get_match_len(m) >= MIN_MATCH ) {
		/* match */
		int l = get_match_len(m);
		int of = get_match_offs(m);
		/*
		 * - does not consider when the escape bit isn't needed.
		 */
		/*
		 * scan through all possible shorter lengths and see if
		 * any are cheaper.
		 * TODO: which choices are actually interesting here?
		 */
		int c = 4;
		while (c && l >= MIN_MATCH) {
		    w = 1 + fast_mlen(l) + fast_moffs(of, l);
		    v = cur + l;

#if DEBUG_COMPLEXITY
		    n_match++;
#endif
		    if (dist[v] > dist[cur] + w) {
			dist[v] = dist[cur] + w;
			prev[v] = cur;
			make_match(&path[v], of, l);
#if DEBUG_COMPLEXITY
			n_match_taken++;
#endif
		    }

		    l--;
		    c--;
		}
	    } else if (is_rle(m) && get_rle_len(m) >= MIN_MATCH) {
		/* match */
		int l = get_rle_len(m);
		int of = 1;
		/*
		 * - does not consider when the escape bit isn't needed.
		 */
		/*
		 * scan through all possible shorter lengths and see if
		 * any are cheaper.
		 * TODO: which choices are actually interesting here?
		 */
		int c = 4;
		while (c && l >= MIN_MATCH) {
		    w = 1 + fast_mlen(l) + fast_moffs(of, l);
		    v = cur + l;

#if DEBUG_COMPLEXITY
		    n_rle++;
#endif
		    if (dist[v] > dist[cur] + w) {
			dist[v] = dist[cur] + w;
			prev[v] = cur;
			make_match(&path[v], of, l);
#if DEBUG_COMPLEXITY
			n_rle_taken++;
#endif
		    }

		    l--;
		    c--;
		}
	    }

	    m++;
	}

	/* literal */
	/*
	 * - should be expanded to handle all possible literal
	 *   sequences
	 */
	w = fast_lit(1);
	v = cur + 1;
#if DEBUG_COMPLEXITY
	n_lt++;
#endif
	if (dist[v] > dist[cur] + w) {
	    dist[v] = dist[cur] + w;
	    prev[v] = cur;
	    make_literal(&path[v], cur, 1);
#if DEBUG_COMPLEXITY
	    n_lt_taken++;
#endif
	}

	cur++;
    }

    msg(MSG_DEBUG, "cost=%d bits (%d bytes)\n", dist[len], (dist[len]+7)/8);

#if DEBUG_COMPLEXITY
    printf("n_match=%d (%d), n_rle=%d (%d), n_lit=%d (%d)\n", n_match, n_match_taken, n_rle, n_rle_taken, n_lt, n_lt_taken);
#endif

    /*
     * Backtrack the cheapest path to find the number of primary units
     * to allocate.
     */
    i = len;
    j = 0;
    while (i > 0) {
	j++;
	i = prev[i];
    }

    PrimaryPath *pp;
    pp = create_primarypath(j, mt->buf);
    pp->len = len;
    pp->cost = dist[len];

    /*
     * Backtrack the cheapest path and create primary units accordingly.
     */
    i = len;
    j = pp->n-1;
    while (i > 0) {
	pp->path[j] = path[i];
	j--;
	i = prev[i];
    }


    /* free up path tables */
    free(dist);
    free(prev);
    free(path);

    return pp;
}


/* eof */
